@@ -16,10 +16,11 @@ import os
 import os.path
 import pathlib
 import pickle
+import warnings
 from dataclasses import dataclass
 from functools import partial
 from io import UnsupportedOperation
-from typing import List, Iterator, TypeVar, Any, Optional
+from typing import List, Iterator, TypeVar, Any, Optional, Callable, AsyncIterator
 
 
 logger = logging.getLogger(__name__)
@@ -51,8 +52,68 @@ def nogc(func, *args, **kwargs):
             gc.enable()
 
 
+def _execute_in_thread_pool(jobs, concurrency: int = None):
+    if concurrency is None:
+        concurrency = 4
+    else:
+        assert 0 <= concurrency <= 16
+        if concurrency < 1:
+            concurrency = 1
+
+    pool = concurrent.futures.ThreadPoolExecutor(concurrency)
+    tasks = []
+    for f, args, kwargs in jobs:
+        tasks.append(pool.submit(f, *args, **kwargs))
+    results = []
+    for f in concurrent.futures.as_completed(tasks):
+        results.append(f.result())
+    return results
+
+
+def _should_update(source: Upath, target: Upath) -> bool:
+    sourceinfo = source.file_info()
+    targetinfo = target.file_info()
+    return (sourceinfo.size != targetinfo.size  # type: ignore
+            or sourceinfo.mtime > targetinfo.mtime)  # type: ignore
+    # Otherwise, we're assuming that
+    # the target file was copied from the source
+    # previously.
+
+
+def _deprecate(oldname, newname):
+    msg = f"`{oldname}` is deprecated; please use `{newname}` instead"
+    warnings.warn(msg, DeprecationWarning)
+
+
 class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
-    _executor: concurrent.futures.ThreadPoolExecutor = None
+    @staticmethod
+    def _should_overwrite(source: Upath,
+                          target: Upath,
+                          *,
+                          exist_action: str = None,
+                          update_filter: Callable[[Upath, Upath], bool] = None,
+                          ) -> bool:
+        # Determine whether `source` should overwrite `target`,
+        # both of which being existing files, usually of the same name
+        # in the context of 'copying' or 'exporting'.
+        if exist_action is None:
+            exist_action = 'raise'
+        else:
+            assert exist_action in ('raise', 'skip', 'overwrite', 'update')
+
+        if exist_action == 'raise':
+            raise FileExistsError(target)
+        if exist_action == 'skip':
+            logger.info(f"target file {target!r} exists; skipped")
+            return False
+        if exist_action == 'update':
+            if update_filter is None:
+                update_filter = _should_update
+            if not update_filter(source, target):
+                logger.info(
+                    f"target {target!r} appears to be up-to-date; skipped")
+                return False
+        return True
 
     def __init__(self, *pathsegments: str, **kwargs):
         '''`Upath` is the base class for a client to a blob store,
@@ -96,27 +157,27 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
         return self._path
 
     def __eq__(self, other) -> bool:
-        if (other.__class__ is not self.__class__):
+        if other.__class__ is not self.__class__:
             return NotImplemented
         return self._path == other._path
 
     def __lt__(self, other) -> bool:
-        if (other.__class__ is not self.__class__):
+        if other.__class__ is not self.__class__:
             return NotImplemented
         return self._path < other._path
 
     def __le__(self, other) -> bool:
-        if (other.__class__ is not self.__class__):
+        if other.__class__ is not self.__class__:
             return NotImplemented
         return self._path <= other._path
 
     def __gt__(self, other) -> bool:
-        if (other.__class__ is not self.__class__):
+        if other.__class__ is not self.__class__:
             return NotImplemented
         return self._path > other._path
 
     def __ge__(self, other) -> bool:
-        if (other.__class__ is not self.__class__):
+        if other.__class__ is not self.__class__:
             return NotImplemented
         return self._path >= other._path
 
@@ -126,102 +187,55 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
     def __truediv__(self: T, key: str) -> T:
         return self.joinpath(key)
 
-    def copy(self: T,
-             target: str,
-             *,
-             concurrency: int = None,
-             exist_action: str = None,
-             ) -> int:
-        '''Copy the content of the current path to the location
-        `target` in the same store. The path `target` is absolute
-        or relative to `self`.
-
-        Return the number of files copied.
-
-        Examples: suppose these blobs are present
-
-            /a/b/c.txt
-            /a/b/c/d/c.txt
-            /e/f/g/xy.data
-            /e/f/g/h/d/dd.txt
-
-        then with `exist_action='raise'`:
-
-            self         target         outcome
-            /a/b/c.txt   ../c           /a/b/c/c.txt
-            /a/b/c.txt   ../c/d         FileExistsError
-            /a/b/c.txt   ../c.txt.back  /a/b/c.txt.back
-            /a/b/c/d     /e/f           /e/f/d/c.txt
-            /a/b/c/d     /e/f/g/h       /e/f/g/h/d/c.txt
-        '''
-        return self.copy_to(self / target,
-                            concurrency=concurrency,
-                            exist_action=exist_action,
-                            )
-
-    def copy_from(self,
-                  source: Upath,
-                  *,
-                  concurrency: int = None,
-                  exist_action: str = None) -> int:
-        '''Copy the content of `source` into `self`.
-
-        `source` may be a file or a directory.
-        In the latter case, its content will be copied recursively.
+    def copy_dir(self: T,
+                 target: str,
+                 *,
+                 concurrency: int = None,
+                 exist_action: str = None,
+                 update_filter: Callable[[Upath, Upath], bool] = None,
+                 ) -> int:
+        '''Analogous to `copy_file`.
 
         Return the number of files copied.
         '''
-        return source.copy_to(self,
-                              concurrency=concurrency,
-                              exist_action=exist_action)
-
-    def _copy_file_to(self, target: Upath, *, exist_action: str) -> int:
-        if target == self:
+        target_ = self.parent / target
+        if target_ == self:
             return 0
 
-        if target.isfile():
-            if exist_action == 'raise':
-                raise FileExistsError(target)
-            if exist_action == 'skip':
-                logger.info(f"target {target!r} exists; skipped")
-                return 0
-            if exist_action == 'update':
-                sourceinfo = self.file_info()
-                targetinfo = target.file_info()
-                if (targetinfo.size == sourceinfo.size
-                        and targetinfo.mtime >= sourceinfo.mtime):
-                    # We're assuming that this suggests
-                    # the target file was copied from the source
-                    # previously.
-                    logger.info(
-                        f"target {target!r} appears to be up-to-date; skipped")
-                    return 0
-            logger.info("copying '%s' to '%s'", self, target)
-            target.write_bytes(self.read_bytes(), overwrite=True)
-            return 1
+        def foo():
+            for p in self.riterdir():
+                extra = str(p.path.relative_to(self.path))
+                yield (
+                    p.copy_file,
+                    [(target_ / extra)._path],
+                    {'exist_action': exist_action, 'update_filter': update_filter},
+                )
 
-        if target.isdir():
-            # Do not delete.
-            raise FileExistsError(target)
+        nn = _execute_in_thread_pool(foo(), concurrency)
+        return sum(nn)
 
-        logger.info("copying '%s' to '%s'", self, target)
-        target.write_bytes(self.read_bytes(), overwrite=False)
-        return 1
+    def _copy_file(self: T, target: T) -> None:
+        # `target` is a path in the same store, but does not exist.
+        # Reference implementation.
+        # Subclass may customize this to perform file operations.
+        target.write_bytes(self.read_bytes())
 
-    def copy_to(self,
-                target: Upath,
-                *,
-                concurrency: int = None,
-                exist_action: str = None,
-                ) -> int:
-        '''Copy the content of `self` to the specified `target`,
-        which is typically in another store.
+    def copy_file(self: T,
+                  target: str,
+                  *,
+                  exist_action: str = None,
+                  update_filter: Callable[[Upath, Upath], bool] = None,
+                  ) -> int:
+        '''Copy file to `target` in the same store.
 
-        `concurrency`: number of threads to use. If `None`,
-        a default value (e.g. 4) is used.
+        `target` is either absolute, or relative to `self.parent`.
+        For example, if `self` is '/a/b/c/d.txt', then
+        `target='e.txt'` means '/a/b/c/e.txt'.
 
-        `exist_action`: what to do when the target file already exists.
-        These are the possible values:
+        If `self` is not an existing file, raise `FileNotFoundError`.
+
+        If `target` is an existing file, then the behavior depends on
+        `exist_action`:
 
             'raise' (default): raise `FileExistsError`.
             'skip': skip this file; proceed to work on other files.
@@ -229,66 +243,34 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
             'update': overwrite if source `mtime` is newer than target,
                 or source and target have diff size; otherwise skip.
 
-        The behavior is analogous to the command `cp` in Linux:
+        If `target` is an existing directory, raise `IsADirectoryError`.
+        Note: this behavior is different from the Unix command `cp`
+        in this situation---it does not *copy into* the target directory.
 
-            # self   target  outcome
-            abc.txt, xy      ==> xy
-            abc.txt, xy/     => xy/abc.txt
-            abc/,    xy      ==> xy/
-            abc/,    xy/     ==> xy/abc/
-
-        Return the number of files copied.
-
-        Compare with `copy`, which copies to another location
-        in the same store.
-
-        Subclasses should provide more efficient implementations
-        if possible, while maintains the behavior defined in
-        this implementation.
+        Return number of files copied (0 or 1).
         '''
-        if exist_action is None:
-            exist_action = 'raise'
-        else:
-            assert exist_action in ('raise', 'skip', 'overwrite', 'update')
+        # Reference implementation.
+        # Subclass should implement by direct file operation if possible.
+        if not self.is_file():
+            raise FileNotFoundError(self)
+        target_ = self.parent / target
+        if target_ == self:
+            return 0
 
-        if target.isdir():
-            target = target / self.name
+        if target_.is_file():
+            if self._should_overwrite(self, target_,
+                                      exist_action=exist_action,
+                                      update_filter=update_filter):
+                target_.remove_file()
+                self._copy_file(target_)
+                return 1
+            return 0
 
-        if self.isfile():
-            return self._copy_file_to(target, exist_action=exist_action)
+        if target_.is_dir():
+            raise IsADirectoryError(target_)
 
-        if self.isdir():
-            if concurrency is None:
-                concurrency = 4
-            else:
-                assert 0 <= concurrency <= 16
-
-            if concurrency <= 1:
-                n = 0
-                for p in self.riterdir():
-                    extra = str(p.path.relative_to(self.path))
-                    k = p._copy_file_to(
-                        target=target/extra,
-                        exist_action=exist_action,
-                    )
-                    n += k
-                return n
-
-            pool = concurrent.futures.ThreadPoolExecutor(concurrency)
-            tasks = []
-            for p in self.riterdir():
-                extra = str(p.path.relative_to(self.path))
-                tasks.append(pool.submit(
-                    p._copy_file_to,
-                    target=target/extra,
-                    exist_action=exist_action
-                ))
-            n = 0
-            for f in concurrent.futures.as_completed(tasks):
-                n += f.result()
-            return n
-
-        raise FileNotFoundError(self)
+        self._copy_file(target_)
+        return 1
 
     def exists(self) -> bool:
         '''Return `True` if the path is an existing file or dir,
@@ -304,17 +286,128 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
         '/a/b' exists, and is a dir;
         '/a/b/c' does not exist.
         '''
-        return self.isfile() or self.isdir()
+        return self.is_file() or self.is_dir()
+
+    def export_dir(self,
+                   target: Upath,
+                   *,
+                   concurrency: int = None,
+                   exist_action: str = None,
+                   update_filter: Callable[[Upath, Upath], bool] = None,
+                   ) -> int:
+        '''Copy the content of `self` to the specified `target`,
+        which is typically in another store.
+
+        Compare with `copy_dir`, which make copies within the same store.
+
+        `concurrency`: number of threads to use. If `None`,
+        a default value (e.g. 4) is used.
+
+        Overwriting happens file-wise. For example, if the target directory
+        contains files that do not exist in the source directory, they
+        are left untouched.
+
+        Return the number of files copied.
+        '''
+        def foo():
+            for p in self.riterdir():
+                extra = str(p.path.relative_to(self.path))
+                yield (
+                    p.export_file,
+                    [target / extra],
+                    {'exist_action': exist_action, 'update_filter': update_filter},
+                )
+
+        nn = _execute_in_thread_pool(foo(), concurrency)
+        return sum(nn)
+
+    def _export_file(self, target: Upath) -> None:
+        # `target` is a non-existent path in another store.
+        # Reference implementation.
+        # Subclass may customize this to perform file download
+        # when `target` is a `LocalUpath`.
+        target.write_bytes(self.read_bytes())
+
+    def export_file(self,
+                    target: Upath,
+                    *,
+                    exist_action: str = None,
+                    update_filter: Callable[[Upath, Upath], bool] = None,
+                    ) -> int:
+        '''Copy the file to the specified `target`, which is typically
+        in another store.
+
+        Return the number of files copied (0 or 1).
+
+        The `target` specifies the name corresponding the the name of `self`.
+        If `target` is an existing directory, a `FileExistsError` is raised.
+        A copy is not placed *into* the target directory. This behavior
+        differs from the Linux command `cp`.
+
+        Compare with `copy_file`, which make copies within the same store.
+        '''
+        if not self.is_file():
+            raise FileNotFoundError(self)
+
+        if target.is_file():
+            if self._should_overwrite(self, target,
+                                      exist_action=exist_action,
+                                      update_filter=update_filter):
+                target.remove_file()
+                logger.info("copying '%s' to '%s'", self, target)
+                self._export_file(target)
+                return 1
+            return 0
+
+        if target.is_dir():
+            # Do not delete.
+            raise FileExistsError(target)
+
+        logger.info("copying '%s' to '%s'", self, target)
+        self._export_file(target)
+        return 1
 
     @ abc.abstractmethod
     def file_info(self) -> Optional[FileInfo]:
         '''
-        If `self.isfile()` is `False`, return `None`.
+        If `self.is_file()` is `False`, return `None`.
         '''
         raise NotImplementedError
 
+    def import_dir(self,
+                   source: Upath,
+                   *,
+                   concurrency: int = None,
+                   exist_action: str = None,
+                   update_filter: Callable[[Upath, Upath], bool] = None,
+                   ) -> int:
+        '''Analogous to `export_dir`.
+        '''
+        return source.export_dir(self,
+                                 concurrency=concurrency,
+                                 exist_action=exist_action,
+                                 update_filter=update_filter)
+
+    def _import_file(self, source: Upath) -> None:
+        # `self` does not exist, hence no concern about overwriting.
+        # Subclass may customize this to perform file upload
+        # when `target` is a `LocalUpath`.
+        # This is not used by `import_file` directly, but
+        # it is used by `export_file` in certain situations.
+        # See `LocalUpath._export_file`.
+        self.write_bytes(source.read_bytes())
+
+    def import_file(self, source: Upath, *,
+                    exist_action: str = None,
+                    update_filter: Callable[[Upath, Upath], bool] = None,
+                    ) -> int:
+        return source.export_file(self,
+                                  exist_action=exist_action,
+                                  update_filter=update_filter,
+                                  )
+
     @ abc.abstractmethod
-    def isdir(self) -> bool:
+    def is_dir(self) -> bool:
         '''Return `True` if the path is an existing directory, `False` otherwise.
 
         If there exists a file named like
@@ -333,13 +426,9 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
         there exist files named `/a/b/...`.
 
         In a local file system, there can be empty directories.
-        However, the class `LocalUpath` tries to prevent that from happening
-        (but can't strictly enforce it).
-        Therefore, `isdir` can be understood as "is a non-empty dir".
-        Consequently, there is usually no need to check whether a directory is
-        "empty".
+        However, it is recommended to not have empty directories.
 
-        Along similar lines, there is no method for "creating a dir" (like `mkdir`).
+        There is no method for "creating a dir" (like `mkdir`).
         Simply create a file under the dir, and the dir will come into being.
         This is analogous to our treatment to files---we don't "create" a file
         in advance; we simply write to a path, intending it to be a file.
@@ -347,7 +436,7 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
         raise NotImplementedError
 
     @ abc.abstractmethod
-    def isfile(self) -> bool:
+    def is_file(self) -> bool:
         '''Return `True` if the path is an existing file, `False` otherwise.
 
         In a cloud blob store, a path can be both a file and a dir. For
@@ -456,17 +545,108 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
         # Refer to https://docs.python.org/3/library/functions.html#open
         return self.read_bytes().decode(encoding=encoding, errors=errors)
 
-    def rename(self: T, target: str, *, overwrite: bool = False) -> T:
-        '''Rename this file or directory to the given `target`
-        in the same store. The path `target` is absolute
-        or relative to `self`.
+    def remove_dir(self, *, concurrency: int = None) -> int:
+        '''Remove the directory pointed to by `self`,
+        along with all its contents, recursively.
 
-        Return a new `Upath` instance pointing to `target`.
+        Return the number of files removed.
 
-        Behavior is similar to `pathlib.Path.rename`.
+        `concurrency`: number of threads to use. If `None`,
+        a default value (e.g. 4) is used.
+
+        Local upath needs to customize this implementation, because
+        it needs to take care of deleting "empty" subdirectories.
+        '''
+        def foo():
+            for p in self.riterdir():
+                yield p.remove_file, [], {}
+
+        nn = _execute_in_thread_pool(foo(), concurrency)
+        return sum(nn)
+
+    @abc.abstractmethod
+    def remove_file(self) -> int:
+        '''Remove the file pointed to by `self`.
+
+        Return the number of files removed (0 or 1).
+
+        If `self` is not an existing file, return 0.
+        If the file exists but can't be removed, raise an exception.
         '''
         raise NotImplementedError
 
+    def rename_dir(self: T,
+                   target: str,
+                   *,
+                   concurrency: int = None,
+                   exist_action: str = None,
+                   update_filter: Callable[[Upath, Upath], bool] = None,
+                   ) -> T:
+        '''Analogous to `rename_file`.
+
+        Local upath needs to customize this implementation, because
+        it needs to take care to delete empty subdirectories under `self`.
+        '''
+        target_ = self.parent / target
+        if target_ == self:
+            return self
+
+        def foo():
+            for p in self.riterdir():
+                extra = str(p.path.relative_to(self.path))
+                yield (
+                    p.rename_file,
+                    [(target_ / extra)._path],
+                    {'exist_action': exist_action, 'update_filter': update_filter},
+                )
+
+        _ = _execute_in_thread_pool(foo(), concurrency)
+        return target_
+
+    def _rename_file(self: T, target: T):
+        '''Rename `self` to `target`, which is a non-existent path
+        in the same store.
+        '''
+        self._copy_file(target)
+        self.remove_file()
+
+    def rename_file(self: T,
+                    target: str,
+                    *,
+                    exist_action: str = None,
+                    update_filter: Callable[[Upath, Upath], bool] = None,
+                    ) -> Optional[T]:
+        '''Rename the current file to `target` in the same store.
+
+        `target` is relative to `self.parent`. For example, if `self`
+        is '/a/b/c/d.txt', then `target='e.txt'` means '/a/b/c/e.txt'.
+
+        Return an object pointing to the new path if renaming happened.
+
+        If `self` is not an existing file, raise `FileNotFoundError`.
+        '''
+        if not self.is_file():
+            raise FileNotFoundError(self)
+        target_ = self.parent / target
+        if target_ == self:
+            return self
+
+        # TODO: rethink this logic.
+        if target_.is_file():
+            if self._should_overwrite(self, target_,
+                                      exist_action=exist_action,
+                                      update_filter=update_filter):
+                target_.remove_file()
+                self._rename_file(target_)
+                return target_
+            return None  # TODO: rethink
+
+        if target_.is_dir():
+            raise FileExistsError(target_)
+
+        self._rename_file(target_)
+        return target_
+
     @abc.abstractmethod
     def riterdir(self: T) -> Iterator[T]:
         '''Yield files under the current dir recursively.
@@ -482,38 +662,6 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
 
         raise NotImplementedError
 
-    @ abc.abstractmethod
-    def rmdir(self, *, missing_ok: bool = False, concurrency: int = None) -> int:
-        '''Remove the directory pointed to by `self`,
-        along with all its contents, recursively.
-
-        Return the number of files removed.
-
-        `concurrency`: number of threads to use. If `None`,
-        a default value (e.g. 4) is used.
-
-        If `self.exists()` is `False` or `self.isdir()` is `False`,
-        and `missing_ok` is `False`, raise `FileNotFoundError`;
-        otherwise, return 0.
-        '''
-        raise NotImplementedError
-
-    @ abc.abstractmethod
-    def rmfile(self, *, missing_ok: bool = False) -> int:
-        '''Remove the file pointed to by `self`.
-
-        Return the number of files removed (0 or 1).
-
-        If `self.exists()` is `False` or `self.isfile()` is `False`,
-        and `missing_ok` is `False`, raise `FileNotFoundError`;
-        otherwise, return 0.
-
-        If the file is the only element in its parent directory,
-        then the directory is also removed. This is to avoid having
-        empty directories. (See doc of `isdir`.)
-        '''
-        raise NotImplementedError
-
     def rmrf(self, *, concurrency: int = None) -> int:
         '''Analogous to `rm -rf`. Remove the file or dir pointed to
         by `self`.
@@ -533,8 +681,8 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
         '''
         if self._path == '/':
             raise UnsupportedOperation("`rmrf` not allowed on root directory")
-        n1 = self.rmfile(missing_ok=True)
-        n2 = self.rmdir(missing_ok=True, concurrency=concurrency)
+        n1 = self.remove_file()
+        n2 = self.remove_dir(concurrency=concurrency)
         return n1 + n2
 
     @ property
@@ -605,214 +753,85 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
         self.write_bytes(z, overwrite=overwrite)
         return n
 
-    async def _a_do(self, func, *args, **kwargs):
-        func = partial(func, *args, **kwargs)
-        return await asyncio.get_running_loop().run_in_executor(
-            self._executor, func)
-
-    async def a_copy(self, target: str, *,
-                     concurrency: int = None, exist_action: str = None):
-        return await self.a_copy_to(self / target,
-                                    concurrency=concurrency,
-                                    exist_action=exist_action,
-                                    )
-
-    async def a_copy_from(self, source: Upath, *,
-                          concurrency: int = None,
-                          exist_action: str = None):
-        return await source.a_copy_to(self,
-                                      concurrency=concurrency,
-                                      exist_action=exist_action,
-                                      )
-
-    async def _a_copy_file_to(self,
-                              target: Upath, *,
-                              exist_action: str,
-                              semaphore: Optional[asyncio.Semaphore] = None,
-                              ) -> int:
-        if target == self:
-            return 0
-
-        async def foo():
-            if await target.a_isfile():
-                if exist_action == 'raise':
-                    raise FileExistsError(target)
-                if exist_action == 'skip':
-                    logger.info(f"target {target!r} exists; skipped")
-                    return 0
-                if exist_action == 'update':
-                    sourceinfo = await self.a_file_info()
-                    targetinfo = await target.a_file_info()
-                    if (targetinfo.size == sourceinfo.size
-                            and targetinfo.mtime >= sourceinfo.mtime):
-                        # We're assuming that this suggests
-                        # the target file was copied from the source
-                        # previously.
-                        return 0
-                logger.info("copying '%s' to '%s'", self, target)
-                await target.a_write_bytes(
-                    await self.a_read_bytes(), overwrite=True)
-                return 1
-
-            if await target.a_isdir():
-                # Do not delete.
-                raise FileExistsError(target)
-
-            logger.info("copying '%s' to '%s'", self, target)
-            await target.a_write_bytes(
-                await self.a_read_bytes(), overwrite=False)
-            return 1
-
-        if semaphore is None:
-            return await foo()
-
-        async with semaphore:
-            return await foo()
-
-    async def a_copy_to(self,
-                        target: Upath,
-                        *,
-                        concurrency: int = None,
-                        exist_action: str = None,
-                        ) -> int:
-        if exist_action is None:
-            exist_action = 'raise'
-        else:
-            assert exist_action in ('raise', 'skip', 'overwrite', 'update')
-
-        if await target.a_isdir():
-            target = target / self.name
-
-        if await self.a_isfile():
-            return await self._a_copy_file_to(
-                target, exist_action=exist_action)
-
-        if not await self.a_isdir():
-            raise FileNotFoundError(self)
-
-        if concurrency is None:
-            concurrency = 4
-        else:
-            assert 0 <= concurrency <= 16
-
-        if concurrency <= 1:
-            n = 0
-            async for p in self.a_riterdir():
-                extra = str(p.path.relative_to(self.path))
-                k = await p._a_copy_file_to(
-                    target/extra,
-                    exist_action=exist_action,
-                )
-                n += k
-            return n
-
-        sem = asyncio.Semaphore(concurrency)
-        tasks = []
-        async for p in self.a_riterdir():
-            extra = str(p.path.relative_to(self.path))
-            tasks.append(p._a_copy_file_to(
-                target/extra,
-                exist_action=exist_action,
-                semaphore=sem,
-            ),
-            )
-        nn = await asyncio.gather(*tasks)
-        return sum(nn)
-
-    async def a_exists(self):
-        return (await self.a_isfile()) or (await self.a_isdir())
-
-    async def a_file_info(self):
-        return await self._a_do(self.file_info)
-
-    async def a_isdir(self):
-        return await self._a_do(self.isdir)
-
-    async def a_isfile(self):
-        return await self._a_do(self.isfile)
-
-    async def a_iterdir(self: T) -> Iterator[T]:
+    async def a_iterdir(self: T) -> AsyncIterator[T]:
         # TODO: may need reimplementation.
         for p in self.iterdir():
             yield p
 
-    @contextlib.asynccontextmanager
-    async def a_lock(self, *, wait: float = 60):
-        # TODO: may need reimplementation.
-        with self.lock(wait=wait):
-            yield
-
-    async def a_ls(self):
-        return sorted((p async for p in self.a_iterdir()))
-
-    async def a_read_bytes(self):
-        return await self._a_do(self.read_bytes)
-
-    async def a_read_json(self, *, no_gc: bool = True, **kwargs):
-        z = await self.a_read_text(**kwargs)
-        if no_gc:
-            return nogc(json.loads, z)
-        return json.loads(z)
-
-    async def a_read_pickle(self, *, no_gc: bool = True):
-        z = await self.a_read_bytes()
-        if no_gc:
-            return nogc(pickle.loads, z)
-        return pickle.loads(z)
-
-    async def a_read_text(self, *,
-                          encoding: str = 'utf-8', errors: str = 'strict'):
-        return (await self.a_read_bytes()).decode(
-            encoding=encoding, errors=errors)
-
-    async def a_rename(self, *args, **kwargs):
-        return await self._a_do(self.rename, *args, **kwargs)
-
-    async def a_riterdir(self: T) -> Iterator[T]:
+    async def a_riterdir(self: T) -> AsyncIterator[T]:
         # TODO: may need reimplementation.
         for p in self.riterdir():
             yield p
 
-    async def a_rmdir(self, *,
-                      missing_ok: bool = False, concurrency: int = None) -> int:
-        # TODO: may need reimplementation.
-        return await self._a_do(self.rmdir,
-                                missing_ok=missing_ok,
-                                concurrency=concurrency)
+    # deprecated on 7/19
 
-    async def a_rmfile(self, *args, **kwargs):
-        return await self._a_do(self.rmfile, *args, **kwargs)
+    def isdir(self):
+        _deprecate('isdir', 'is_dir')
+        return self.is_dir()
 
-    async def a_rmrf(self, *, concurrency: int = None) -> int:
-        if self._path == '/':
-            raise UnsupportedOperation(
-                "`a_rmrf` not allowed on root directory")
-        n1 = await self.a_rmfile(missing_ok=True)
-        n2 = await self.a_rmdir(missing_ok=True, concurrency=concurrency)
-        return n1 + n2
+    def isfile(self):
+        _deprecate('isfile', 'is_file')
+        return self.is_file()
 
-    async def a_write_bytes(self, *args, **kwargs):
-        return await self._a_do(self.write_bytes, *args, **kwargs)
+    async def a_isdir(self):
+        _deprecate('a_isdir', 'a_is_dir')
+        return await self.a_is_dir()
 
-    async def a_write_json(self, data, *, overwrite=False, **kwargs) -> int:
-        return await self.a_write_text(
-            json.dumps(data), overwrite=overwrite, **kwargs)
+    async def a_isfile(self):
+        _deprecate('a_isfile', 'a_is_file')
+        return await self.a_is_file()
 
-    async def a_write_pickle(self, data, *, overwrite=False) -> int:
-        return await self.a_write_bytes(
-            pickle.dumps(data, protocol=pickle.HIGHEST_PROTOCOL),
-            overwrite=overwrite,
+    def rmdir(self, *, missing_ok=False, concurrency=None):
+        _deprecate('rmdir', 'remove_dir')
+        n = self.remove_dir(concurrency=concurrency)
+        if n == 0 and not missing_ok:
+            raise FileNotFoundError(self)
+        return n
+
+    def rmfile(self, *, missing_ok=False):
+        _deprecate('rmfile', 'remove_file')
+        n = self.remove_file()
+        if n == 0 and not missing_ok:
+            raise FileNotFoundError(self)
+        return n
+
+    async def a_rmdir(self, *, missing_ok=False, concurrency=None):
+        _deprecate('a_rmdir', 'a_remove_dir')
+        n = await self.a_remove_dir(concurrency=concurrency)
+        if n == 0 and not missing_ok:
+            raise FileNotFoundError(self)
+        return n
+
+    async def a_rmfile(self, *, missing_ok=False):
+        _deprecate('a_rmfile', 'a_remove_file')
+        n = await self.a_remove_file()
+        if n == 0 and not missing_ok:
+            raise FileNotFoundError(self)
+        return n
+
+
+def make_a_method(name):
+    async def f(self, *args, **kwargs):
+        f = partial(getattr(self, name), *args, **kwargs)
+        return await asyncio.get_running_loop().run_in_executor(
+            None, f
         )
 
-    async def a_write_text(
-            self,
-            data: str,
-            *,
-            overwrite: bool = False,
-            encoding: str = 'utf-8',
-            errors: str = 'strict',
-    ) -> int:
-        n = len(data)
-        z = data.encode(encoding=encoding, errors=errors)
-        await self.a_write_bytes(z, overwrite=overwrite)
-        return n
+    f.__name__ = f'a_{name}'
+    return f
+
+
+for m in ('copy_dir', 'copy_file',
+          'export_dir', 'export_file',
+          'exists',
+          'file_info',
+          'import_dir', 'import_file',
+          'is_dir', 'is_file',
+          'ls',
+          'read_bytes', 'read_json', 'read_pickle', 'read_text',
+          'remove_dir', 'remove_file',
+          'rename_dir', 'rename_file',
+          'rmrf',
+          'write_bytes', 'write_json', 'write_pickle', 'write_text',
+          ):
+    setattr(Upath, f'a_{m}', make_a_method(m))
